//############################################################################
// ~ International Laser Display Association ~
// ~     ILDA File Format Driver in C++      ~
// ~             Built for EZFB              ~
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#ifndef __EZ_ILDA_UTILITY_DEFINITIONS__
#define __EZ_ILDA_UTILITY_DEFINITIONS__

//############################################################################
#include "ez_ilda_point.hpp"

//############################################################################
void    undo_wave_temp(ifstream& in, ofstream& out, string file_name);

//############################################################################
int     distance_points(const ez_ilda_point p1, const ez_ilda_point p2);
int     dwell_points   (double angle);
u_int   next_hue       ();

//############################################################################
inline ez_3D_float rotate_point(ez_3D_float p, ez_3D_float a)
{
    ez_3D_float   sin_a(sin(a.x), sin(a.y), sin(a.z)),
                  cos_a(cos(a.x), cos(a.y), cos(a.z)),
                  rotated;
                    
    rotated.y =       p.y * cos_a.x -  p.z * sin_a.x;
    rotated.z =       p.z * cos_a.x +  p.y * sin_a.x; // rotate around x
    rotated.z = rotated.z * cos_a.y -  p.x * sin_a.y;
    rotated.x =       p.x * cos_a.y + (p.z * cos_a.x +  p.y * sin_a.x) * sin_a.y; // rotate around y
    rotated.x = rotated.x * cos_a.z - (p.y * cos_a.x -  p.z * sin_a.x) * sin_a.z;
    rotated.y = rotated.y * cos_a.z + (p.x * cos_a.y + (p.z * cos_a.x +  p.y * sin_a.x) * sin_a.y) * sin_a.z; // rotate around z
    return rotated;
}

//############################################################################
inline ez_3D_float rotate_point_around_x(ez_3D_float p, float a)
{
    float         sin_a(sin(a)),
                  cos_a(cos(a));
    ez_3D_float   rotated(p);

    rotated.y = p.y * cos_a -  p.z * sin_a;
    rotated.z = p.z * cos_a +  p.y * sin_a; // rotate around x
    return rotated;
}

//############################################################################
inline ez_3D_float rotate_point_around_y(ez_3D_float p, float a)
{
    float         sin_a(sin(a)),
                  cos_a(cos(a));
    ez_3D_float   rotated(p);

    rotated.x = p.x * cos_a -  p.z * sin_a;
    rotated.z = p.z * cos_a +  p.x * sin_a; // rotate around y
    return rotated;
}

//############################################################################
inline ez_3D_float rotate_point_around_z(ez_3D_float p, float a)
{
    float         sin_a(sin(a)),
                  cos_a(cos(a));
    ez_3D_float   rotated(p);

    rotated.x = p.x * cos_a -  p.y * sin_a;
    rotated.y = p.y * cos_a +  p.x * sin_a; // rotate around z
    return rotated;
}

//############################################################################
inline ez_3D_float rotate_point_around_point(ez_3D_float p1, ez_3D_float p2, ez_3D_float a)
{
    return rotate_point(p1 - p2, a) + p2;
}

//############################################################################
inline ez_3D_float rotate_point_around_point_x(ez_3D_float p1, ez_3D_float p2, float a)
{
    return rotate_point_around_x(p1 - p2, a) + p2;
}

//############################################################################
inline ez_3D_float rotate_point_around_point_y(ez_3D_float p1, ez_3D_float p2, float a)
{
    return rotate_point_around_y(p1 - p2, a) + p2;
}

//############################################################################
inline ez_3D_float rotate_point_around_point_z(ez_3D_float p1, ez_3D_float p2, float a)
{
    return rotate_point_around_z(p1 - p2, a) + p2;
}

//############################################################################
inline ez_3D_float scale_point_around_point(ez_3D_float p1, ez_3D_float  p2, ez_3D_float m)
{
    return (m * (p1 - p2)) + p2;
}

//############################################################################
inline int ez_ilda_bounds_check( const ez_3D_float& f,
                                 const ez_3D_short& max,
                                 const ez_3D_short& min
                               )
{
    if(f.x > max.x)  return EZ_OUT_POS_X;
    if(f.x < min.x)  return EZ_OUT_NEG_X;
    if(f.y > max.y)  return EZ_OUT_POS_Y;
    if(f.y < min.y)  return EZ_OUT_NEG_Y;
    if(f.z > max.z)  return EZ_OUT_POS_Z;
    if(f.z < min.z)  return EZ_OUT_NEG_Z;
    return EZ_IN_BOUNDS;
}

//############################################################################
inline int ez_ilda_bounds_check(const ez_3D_float& f, const ez_ilda_3D_space& space)
{
    ez_3D_short max, min;
    switch(space)
    {
        case EZ_ILDA_CUBE          : max =  32767 ;
                                     min = -32768 ;
                                     break        ;
        //--------------------------------------------------------------------
        case EZ_ILDA_SPHERE        : max =  23169 ;
                                     min = -23169 ;
                                     break        ;
        //--------------------------------------------------------------------
        case EZ_ILDA_7_8THS_CUBE   : max =  28672 ;
                                     min = -28672 ;
                                     break        ;
        //--------------------------------------------------------------------
        case EZ_ILDA_7_8THS_SPHERE : max =  20274 ;
                                     min = -20274 ;
                                     break        ;
        //--------------------------------------------------------------------
    }
    return ez_ilda_bounds_check(f, max, min);
}

//############################################################################
#endif

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
