//############################################################################
// ~ International Laser Display Association ~
// ~     ILDA File Format Driver in C++      ~
// ~             Built for EZFB              ~
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#ifndef __EZ_PALETTE_VECTOR_DEFINITIONS__
#define __EZ_PALETTE_VECTOR_DEFINITIONS__

//############################################################################
#include "ez_palette.hpp"

//############################################################################
class ez_palette_set : public vector<ez_palette>
{
public:
    ez_palette_set()
                  : palette_index          (EZ_ILDA_DEFAULT),
                    target_palette_index   (0              ),
                    selected_palette_color (0              ) 
                       {    populate_palette_set();    }
    //------------------------------------------------------------------------
   ~ez_palette_set()   {    clear();    }
    //------------------------------------------------------------------------
    ez_palette_set& replace_palette_set(const ez_palette_set& palette_set)
                    {
                        clear();
                        insert(begin(), palette_set.begin(), palette_set.end());
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_palette_set& operator = (const ez_palette_set& palette_set)
                    {
                        clear();
                        insert(begin(), palette_set.begin(), palette_set.end());
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_palette_set& operator += (const ez_palette_set& palette_set)
                    {
                        insert(end(), palette_set.begin(), palette_set.end());
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_palette_set& operator += (const ez_palette& palette)
                    {
                        push_back(palette);
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_palette_set& push_back_palette(const ez_palette& palette)
                    {
                        push_back(palette);
                        return *this;
                    }
    //------------------------------------------------------------------------
    void  set_palette_index(int index) //sets palette_index
            {
                if(number_of_palettes())
                {
                    if(index >= number_of_palettes())
                        index = 0;
                    if(index < 0)
                        index = number_of_palettes() - 1;
                    palette_index = index;
                }
                else
                    palette_index = 0;
            }
    //------------------------------------------------------------------------
    void  set_target_palette_index(int index) //sets palette_index
            {
                if(number_of_palettes())
                {
                    if(index >= number_of_palettes())
                        index = 0;
                    if(index < 0)
                        index = number_of_palettes() - 1;
                    target_palette_index = index;
                }
                else
                    target_palette_index = 0;
            }
    //------------------------------------------------------------------------
    ez_palette&  palette(int index) //sets palette_index
                    {
                        if(number_of_palettes())
                        {
                            if(index >= number_of_palettes())
                                index = 0;
                            if(index < 0)
                                index = number_of_palettes() - 1;
                            palette_index = index;
                            return at(palette_index);
                        }
                        return front();
                    }
    //------------------------------------------------------------------------
    ez_palette&  target_palette(int index) //sets target_palette_index
                    {
                        if(number_of_palettes())
                        {
                            if(index >= number_of_palettes())
                                index = 0;
                            if(index < 0)
                                index = number_of_palettes() - 1;
                            target_palette_index = index;
                            return at(target_palette_index);
                        }
                        return front();
                    }
    //------------------------------------------------------------------------
    ez_palette& current_palette()
                    {
                        if(palette_index != EZ_TRUE_COLOR)
                            return at(palette_index);
                        return at(target_palette_index);
                    }
    //------------------------------------------------------------------------
    void delete_current_palette()
                {
                    if(number_of_palettes())
                        erase(begin() + palette_index);
                    return;
                }
    //------------------------------------------------------------------------
    int  add_palette_from_file(string file_name)
                {
                    ez_palette palette;
                    if(palette.from_ascii_file(file_name) == EZ_ILDA_OK)
                    {
                        push_back(palette);
                        target_palette_index = number_of_palettes() - 1;
                        return EZ_ILDA_OK;
                    }
                    return EZ_ILDA_FILE_OPEN_FAILED;
                }
    //------------------------------------------------------------------------
    int   number_of_palettes            () const {   return (int)size();   }
    ez_palette   palette_picker                (int index) const;
    void         render_palette_in_ezfb_bitmap () const   ;
    void         populate_palette_set          ()         ;
    void         display_stats                 () const   ;
    void         set_selected_palette_color    (u_char c) ;
    //------------------------------------------------------------------------
    int          palette_index          ,
                 target_palette_index   ,
                 selected_palette_color ;
    //------------------------------------------------------------------------
};

//############################################################################
#endif

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
