//############################################################################
// ~ International Laser Display Association ~
// ~     ILDA file format driver in C++      ~
// ~             built for EZFB              ~
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#include "ez_wave.hpp"
    
//############################################################################
ez_wave_header::ez_wave_header( const int    s,
                                const int    r,
                                const short  n,
                                const short  b
                              )
              : num_samples     (s),
                sample_rate     (r),
                ez_wave_mode    (   EZ_WAVE_POSITIVE
                                  | EZ_WAVE_END_OF_FRAME
                                  | EZ_WAVE_UNIQUE_FRAME
                                )  ,
                num_channels    (n),
                bits_per_sample (b),
                header_version  (LASERBOY_WAVE_VERSION),
                parms           ()
{
    offset = new int[num_channels];
    for(int i = 0; i < num_channels; i++)
        offset[i] = 0;
}

//############################################################################
ez_wave_header::~ez_wave_header()
{
    delete offset;
}

//############################################################################
ez_wave_header::ez_wave_header(const ez_wave_header& header)
              : num_samples     (header.num_samples    ),
                sample_rate     (header.sample_rate    ),
                ez_wave_mode    (header.ez_wave_mode   ),
                num_channels    (header.num_channels   ),
                bits_per_sample (header.bits_per_sample)
{    
    offset = new int[header.num_channels];
    memcpy((void*)offset, (void*)header.offset, header.num_channels * sizeof(int));
    if(header.ez_wave_mode & EZ_WAVE_OPTIMIZED)
    {
        parms.lit_dwell_overhang     = header.parms.lit_dwell_overhang    ;
        parms.distance_delta_max     = header.parms.distance_delta_max    ;
        parms.longest_dwell_microsec = header.parms.longest_dwell_microsec;
        parms.insignificant_distance = header.parms.insignificant_distance;
        parms.zero_angle_tolerance   = header.parms.zero_angle_tolerance  ;
        parms.frames_per_second      = header.parms.frames_per_second     ;
    }
}

//############################################################################
ez_wave_header& ez_wave_header::operator = (const ez_wave_header& header)
{
    num_samples     = header.num_samples    ;
    sample_rate     = header.sample_rate    ;
    ez_wave_mode    = header.ez_wave_mode   ;
    num_channels    = header.num_channels   ;
    bits_per_sample = header.bits_per_sample;
    
    delete offset;
    offset = new int[header.num_channels];
    memcpy((void*)offset, (void*)header.offset, header.num_channels * sizeof(int));
    
    if(header.ez_wave_mode & EZ_WAVE_OPTIMIZED)
    {
        parms.lit_dwell_overhang     = header.parms.lit_dwell_overhang    ;
        parms.distance_delta_max     = header.parms.distance_delta_max    ;
        parms.longest_dwell_microsec = header.parms.longest_dwell_microsec;
        parms.insignificant_distance = header.parms.insignificant_distance;
        parms.zero_angle_tolerance   = header.parms.zero_angle_tolerance  ;
        parms.frames_per_second      = header.parms.frames_per_second     ;
    }
    return *this;
}

//############################################################################
int ez_wave_header::from_ifstream_wave(ifstream& in)
{
    char    a, b, c, d;
    
    short   audio_format    ,
            block_align     ;
            
    int     i               ,
            chunk_size      ,
            sub_chunk_1_size,
            sub_chunk_2_size,
            byte_rate       ;
            
    string  temp_string     ;
    //------------------------------------------------------------------------
    if(!in.eof())
    {
        in.get(a); if(a != 'R') return EZ_ILDA_BAD_HEADER;
        in.get(a); if(a != 'I') return EZ_ILDA_BAD_HEADER;
        in.get(a); if(a != 'F') return EZ_ILDA_BAD_HEADER;
        in.get(a); if(a != 'F') return EZ_ILDA_BAD_HEADER;
        //--------------------------------------------------------------------
        in.get(a); in.get(b); in.get(c); in.get(d);
        chunk_size = (   (a & 0x000000ff)
                       | (b & 0x000000ff) << 8 
                       | (c & 0x000000ff) << 16
                       | (d & 0x000000ff) << 24
                     );
        //--------------------------------------------------------------------
        in.get(a); if(a != 'W') return EZ_ILDA_BAD_HEADER;
        in.get(a); if(a != 'A') return EZ_ILDA_BAD_HEADER;
        in.get(a); if(a != 'V') return EZ_ILDA_BAD_HEADER;
        in.get(a); if(a != 'E') return EZ_ILDA_BAD_HEADER;
        //--------------------------------------------------------------------
        in.get(a); if(a != 'f') return EZ_ILDA_BAD_HEADER;
        in.get(a); if(a != 'm') return EZ_ILDA_BAD_HEADER;
        in.get(a); if(a != 't') return EZ_ILDA_BAD_HEADER;
        in.get(a); if(a != ' ') return EZ_ILDA_BAD_HEADER;
        //--------------------------------------------------------------------
        in.get(a); in.get(b); in.get(c); in.get(d);
        sub_chunk_1_size = (   (a & 0x000000ff)
                             | (b & 0x000000ff) << 8 
                             | (c & 0x000000ff) << 16
                             | (d & 0x000000ff) << 24
                           );
        //--------------------------------------------------------------------
        in.get(a); in.get(b);
        audio_format = ((b & 0x00ff) << 8 | (a & 0x00ff));  
        if(audio_format != 1) return EZ_ILDA_INCOMPATIBLE_FORMATS;
        //--------------------------------------------------------------------
        in.get(a); in.get(b);
        num_channels = ((b & 0x00ff) << 8 | (a & 0x00ff));        
        //--------------------------------------------------------------------
        in.get(a); in.get(b); in.get(c); in.get(d);
        sample_rate = (   (a & 0x000000ff)
                        | (b & 0x000000ff) << 8 
                        | (c & 0x000000ff) << 16
                        | (d & 0x000000ff) << 24
                      );
        //--------------------------------------------------------------------
        in.get(a); in.get(b); in.get(c); in.get(d);
        byte_rate = (   (a & 0x000000ff)
                      | (b & 0x000000ff) << 8 
                      | (c & 0x000000ff) << 16
                      | (d & 0x000000ff) << 24
                    );
        //--------------------------------------------------------------------
        in.get(a); in.get(b);
        block_align = ((b & 0x00ff) << 8 | (a & 0x00ff));
        //--------------------------------------------------------------------
        in.get(a); in.get(b);
        bits_per_sample = ((b & 0x00ff) << 8 | (a & 0x00ff));
        //--------------------------------------------------------------------
        if(byte_rate   != (sample_rate * num_channels * (bits_per_sample / 8)))
            return EZ_ILDA_BAD_HEADER;
        if(block_align != (num_channels * (bits_per_sample / 8)))
            return EZ_ILDA_BAD_HEADER;
        //--------------------------------------------------------------------
        if(sub_chunk_1_size > 16) // something has been added to this header
        {
            //----------------------------------------------------------------
            in.get(a);
            temp_string.push_back(a);
            in.get(a);
            temp_string.push_back(a);
            in.get(a);
            temp_string.push_back(a);
            in.get(a);
            temp_string.push_back(a);
            in.get(a);
            temp_string.push_back(a);
            in.get(a);
            temp_string.push_back(a);
            in.get(a);
            temp_string.push_back(a);
            in.get(a);
            temp_string.push_back(a);
            //----------------------------------------------------------------
            if(temp_string == "offsets ")
            {
                //------------------------------------------------------------
                delete offset;
                offset = new int[num_channels];
                for(i = 0; i < num_channels; i++)
                {
                    in.get(a); in.get(b); in.get(c); in.get(d);
                    offset[i] = (   (a & 0x000000ff)
                                  | (b & 0x000000ff) << 8 
                                  | (c & 0x000000ff) << 16
                                  | (d & 0x000000ff) << 24
                                );
                }
                //------------------------------------------------------------
                in.get(a); in.get(b); in.get(c); in.get(d);
                if(    (a == 'd')
                    && (b == 'a')
                    && (c == 't')
                    && (d == 'a')
                  )
                {
                    header_version = "v0 offsets       ";
                } // end if( we just read "data" )
                //------------------------------------------------------------
                else
                {
                    if(    (a == 'e') // we may or may not have an "eof3" mark
                        && (b == 'o') // if we do we don't have ez_wave_mode
                        && (c == 'f') // ez_wave_mode has default values in the ctor
                        && (d == '3')
                      )
                    {
                        header_version = "v1 offsets eof3  ";
                    } // end if( we just read "eof3" )
                    //--------------------------------------------------------
                    else            
                    {
                        ez_wave_mode = (   (a & 0x000000ff)
                                         | (b & 0x000000ff) << 8 
                                         | (c & 0x000000ff) << 16
                                         | (d & 0x000000ff) << 24
                                       );
                        header_version = "v2 ez_wave_mode  ";
                        
                    } // end else we take this as ez_wave_mode
                    //--------------------------------------------------------
                    ez_wave_mode |= EZ_WAVE_OFFSETS; // we know we have offsets
                    //--------------------------------------------------------
                    in.get(a); in.get(b); in.get(c); in.get(d);
                    //--------------------------------------------------------
                    if(a != 'd') return EZ_ILDA_BAD_HEADER;
                    if(b != 'a') return EZ_ILDA_BAD_HEADER;
                    if(c != 't') return EZ_ILDA_BAD_HEADER;
                    if(d != 'a') return EZ_ILDA_BAD_HEADER;                    
                }                    
            } // end if(temp_string == "offsets ")
            //----------------------------------------------------------------
            else if(temp_string == "LaserBoy")
            {
                in.get(a);
                temp_string.push_back(a);
                in.get(a);
                temp_string.push_back(a);
                in.get(a);
                temp_string.push_back(a);
                in.get(a);
                temp_string.push_back(a);
                in.get(a);
                temp_string.push_back(a);
                in.get(a);
                temp_string.push_back(a);
                in.get(a);
                temp_string.push_back(a);
                in.get(a);
                temp_string.push_back(a);                                
                //------------------------------------------------------------
                if(    temp_string != "LaserBoy06072008"
                    && temp_string != LASERBOY_WAVE_VERSION
                  )
                    return EZ_ILDA_INCOMPATIBLE_FORMATS;
                //------------------------------------------------------------
                header_version = temp_string;
                //------------------------------------------------------------
                in.get(a); in.get(b); in.get(c); in.get(d);
                ez_wave_mode = (   (a & 0x000000ff)
                                 | (b & 0x000000ff) << 8 
                                 | (c & 0x000000ff) << 16
                                 | (d & 0x000000ff) << 24
                               );
                //------------------------------------------------------------
                delete offset;
                offset = new int[num_channels];
                //------------------------------------------------------------
                if(ez_wave_mode & EZ_WAVE_OFFSETS)
                    for(i = 0; i < num_channels; i++)
                    {
                        in.get(a); in.get(b); in.get(c); in.get(d);
                        offset[i] = (   (a & 0x000000ff)
                                      | (b & 0x000000ff) << 8 
                                      | (c & 0x000000ff) << 16
                                      | (d & 0x000000ff) << 24
                                    );
                    }
                //------------------------------------------------------------
                else
                    for(i = 0; i < num_channels; i++)
                        offset[i] = 0;
                //------------------------------------------------------------
                if(ez_wave_mode & EZ_WAVE_OPTIMIZED)
                {
                    in.get(a); in.get(b); in.get(c); in.get(d);
                    parms.lit_dwell_overhang = (   (a & 0x000000ff)
                                                 | (b & 0x000000ff) << 8 
                                                 | (c & 0x000000ff) << 16
                                                 | (d & 0x000000ff) << 24
                                               );
                    //--------------------------------------------------------
                    if(header_version == "LaserBoy06072008")
                    {
                        in.get(a); in.get(b); in.get(c); in.get(d);
                        parms.distance_delta_max = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8 
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(a); in.get(b); in.get(c); in.get(d);
                        parms.longest_dwell_microsec = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8 
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(a); in.get(b); in.get(c); in.get(d);
                        parms.insignificant_distance = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8 
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(a); in.get(b); in.get(c); in.get(d);
                        parms.zero_angle_tolerance = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8 
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(a); in.get(b); in.get(c); in.get(d);
                        parms.frames_per_second = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8 
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                    }
                    //--------------------------------------------------------
                    else
                    {                     
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.distance_delta_max = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8 
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.longest_dwell_microsec = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8 
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.insignificant_distance = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8 
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.zero_angle_tolerance = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8 
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                        in.get(d); in.get(c); in.get(b); in.get(a);
                        parms.frames_per_second = (float)
                                                   (   (a & 0x000000ff)
                                                     | (b & 0x000000ff) << 8 
                                                     | (c & 0x000000ff) << 16
                                                     | (d & 0x000000ff) << 24
                                                   );
                    }
                }
                //------------------------------------------------------------
                in.get(a); in.get(b); in.get(c); in.get(d);
                if(a != 'd') return EZ_ILDA_BAD_HEADER;
                if(b != 'a') return EZ_ILDA_BAD_HEADER;
                if(c != 't') return EZ_ILDA_BAD_HEADER;
                if(d != 'a') return EZ_ILDA_BAD_HEADER;
            //----------------------------------------------------------------
            } // end else if(temp_string == "LaserBoy")
            else
                return EZ_ILDA_INCOMPATIBLE_FORMATS;
        } // end if(sub_chunk_1_size > 16)            
        //--------------------------------------------------------------------
        in.get(a); in.get(b); in.get(c); in.get(d);
        sub_chunk_2_size = (   (a & 0x000000ff)
                             | (b & 0x000000ff) << 8 
                             | (c & 0x000000ff) << 16
                             | (d & 0x000000ff) << 24
                           );
        //if(sub_chunk_2_size != (chunk_size + 36)) return EZ_ILDA_BAD_HEADER;
        //--------------------------------------------------------------------
        num_samples = sub_chunk_2_size / (num_channels * (bits_per_sample / 8));
    }
    return 0;//in.bad(); // this is a question!
}

//############################################################################
void ez_wave_header::to_ofstream_wave(ofstream& out, bool with_laserboy_stuff) const
{
    short   audio_format     = 1,
            block_align      = num_channels * (bits_per_sample / 8);

    int     i,
            sub_chunk_1_size = 16, // standard wave Subchunk1Size                                  
            sub_chunk_2_size = num_samples * num_channels * (bits_per_sample / 8),
            byte_rate        = sample_rate * num_channels * (bits_per_sample / 8),
            chunk_size       = 20 + sub_chunk_1_size + sub_chunk_2_size;
    //------------------------------------------------------------------------
    if(with_laserboy_stuff)
        sub_chunk_1_size += (   16 // sizeof "LaserBoy06062008"
                              +  4 // sizeof (int)ez_wave_mode
                              + (   (ez_wave_mode & EZ_WAVE_OFFSETS)
                                  ? (num_channels * sizeof(int))
                                  : (0)
                                )
                              + (   (ez_wave_mode & EZ_WAVE_OPTIMIZED)
                                  ? (6 * sizeof(int)) // six 32bit numbers
                                  : (0)
                                )
                            );
    //------------------------------------------------------------------------
    out.put('R');
    out.put('I');
    out.put('F');
    out.put('F');

    out.put((char) (chunk_size & 0x000000ff)       );
    out.put((char)((chunk_size & 0x0000ff00) >> 8 ));
    out.put((char)((chunk_size & 0x00ff0000) >> 16));
    out.put((char)((chunk_size & 0xff000000) >> 24));

    out.put('W');
    out.put('A');
    out.put('V');
    out.put('E');

    out.put('f');
    out.put('m');
    out.put('t');
    out.put(' ');

    out.put((char) (sub_chunk_1_size & 0x000000ff)       );
    out.put((char)((sub_chunk_1_size & 0x0000ff00) >> 8 ));
    out.put((char)((sub_chunk_1_size & 0x00ff0000) >> 16));
    out.put((char)((sub_chunk_1_size & 0xff000000) >> 24));

    out.put((char) (audio_format & 0x00ff)      );
    out.put((char)((audio_format & 0xff00) >> 8)); // 2

    out.put((char) (num_channels & 0x00ff)      );
    out.put((char)((num_channels & 0xff00) >> 8)); // 4

    out.put((char) (sample_rate & 0x000000ff)       );
    out.put((char)((sample_rate & 0x0000ff00) >> 8 ));
    out.put((char)((sample_rate & 0x00ff0000) >> 16));
    out.put((char)((sample_rate & 0xff000000) >> 24)); // 8

    out.put((char) (byte_rate & 0x000000ff)       );
    out.put((char)((byte_rate & 0x0000ff00) >> 8 ));
    out.put((char)((byte_rate & 0x00ff0000) >> 16));
    out.put((char)((byte_rate & 0xff000000) >> 24)); // 12

    out.put((char) (block_align & 0x00ff)      );
    out.put((char)((block_align & 0xff00) >> 8)); // 14

    out.put((char) (bits_per_sample & 0x00ff)      );
    out.put((char)((bits_per_sample & 0xff00) >> 8)); // 16
//----------------------------------------------------------------------------
// extra structures added for LaserBoy !!!
//----------------------------------------------------------------------------
    if(with_laserboy_stuff)
    {
        out.put(header_version[ 0]); // L
        out.put(header_version[ 1]); // a
        out.put(header_version[ 2]); // s
        out.put(header_version[ 3]); // e
        out.put(header_version[ 4]); // r
        out.put(header_version[ 5]); // B
        out.put(header_version[ 6]); // o
        out.put(header_version[ 7]); // y
        
        out.put(header_version[ 8]);
        out.put(header_version[ 9]);
        out.put(header_version[10]);
        out.put(header_version[11]);
        out.put(header_version[12]);
        out.put(header_version[13]);
        out.put(header_version[14]);
        out.put(header_version[15]);
        //--------------------------------------------------------------------
        out.put((char) (ez_wave_mode & 0x000000ff)       );
        out.put((char)((ez_wave_mode & 0x0000ff00) >> 8 ));
        out.put((char)((ez_wave_mode & 0x00ff0000) >> 16));
        out.put((char)((ez_wave_mode & 0xff000000) >> 24)); // 36
        //--------------------------------------------------------------------
        if(ez_wave_mode & EZ_WAVE_OFFSETS)
            for(i = 0; i < num_channels; i++)
            {
                out.put((char)( offset[i] & 0x000000ff)       );
                out.put((char)((offset[i] & 0x0000ff00) >> 8 ));
                out.put((char)((offset[i] & 0x00ff0000) >> 16));
                out.put((char)((offset[i] & 0xff000000) >> 24)); // 36 + 4 * num_channels
            }
        //--------------------------------------------------------------------
        if(ez_wave_mode & EZ_WAVE_OPTIMIZED)
        {
            out.put((char) (parms.lit_dwell_overhang & 0x000000ff)       ); // int
            out.put((char)((parms.lit_dwell_overhang & 0x0000ff00) >> 8 ));
            out.put((char)((parms.lit_dwell_overhang & 0x00ff0000) >> 16));
            out.put((char)((parms.lit_dwell_overhang & 0xff000000) >> 24));
                    
            out.put((char)(((int)(parms.distance_delta_max) & 0xff000000) >> 24)); // float
            out.put((char)(((int)(parms.distance_delta_max) & 0x00ff0000) >> 16));
            out.put((char)(((int)(parms.distance_delta_max) & 0x0000ff00) >> 8 ));
            out.put((char) ((int)(parms.distance_delta_max) & 0x000000ff)       );
                    
            out.put((char)(((int)(parms.longest_dwell_microsec) & 0xff000000) >> 24)); // float
            out.put((char)(((int)(parms.longest_dwell_microsec) & 0x00ff0000) >> 16));
            out.put((char)(((int)(parms.longest_dwell_microsec) & 0x0000ff00) >> 8 ));
            out.put((char) ((int)(parms.longest_dwell_microsec) & 0x000000ff)       );
                    
            out.put((char)(((int)(parms.insignificant_distance) & 0xff000000) >> 24)); // float
            out.put((char)(((int)(parms.insignificant_distance) & 0x00ff0000) >> 16));
            out.put((char)(((int)(parms.insignificant_distance) & 0x0000ff00) >> 8 ));
            out.put((char) ((int)(parms.insignificant_distance) & 0x000000ff)       );
                    
            out.put((char)(((int)(parms.zero_angle_tolerance) & 0xff000000) >> 24)); // float
            out.put((char)(((int)(parms.zero_angle_tolerance) & 0x00ff0000) >> 16));
            out.put((char)(((int)(parms.zero_angle_tolerance) & 0x0000ff00) >> 8 ));
            out.put((char) ((int)(parms.zero_angle_tolerance) & 0x000000ff)       );
                    
            out.put((char)(((int)(parms.frames_per_second) & 0xff000000) >> 24)); // float
            out.put((char)(((int)(parms.frames_per_second) & 0x00ff0000) >> 16));
            out.put((char)(((int)(parms.frames_per_second) & 0x0000ff00) >> 8 ));
            out.put((char) ((int)(parms.frames_per_second) & 0x000000ff)       );
        }
    } // end if(with_laserboy_stuff)
    //------------------------------------------------------------------------
    out.put('d');
    out.put('a');
    out.put('t');
    out.put('a');
    
    out.put((char) (sub_chunk_2_size & 0x000000ff)       );
    out.put((char)((sub_chunk_2_size & 0x0000ff00) >> 8 ));
    out.put((char)((sub_chunk_2_size & 0x00ff0000) >> 16));
    out.put((char)((sub_chunk_2_size & 0xff000000) >> 24));
    return;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
